#include "Lighting.h"

// light 1 acts as the sun and provides ambient lighting
GLfloat white[] = { 1.0, 1.0, 1.0, 1.0 };
GLfloat ltgray[] = { 0.8, 0.8, 0.8, 1.0 };
GLfloat mdgray[] = { 0.5, 0.5, 0.5, 1.0 };
GLfloat dkgray[] = { 0.3, 0.3, 0.3, 1.0 };
GLfloat black[] = { 0.0, 0.0, 0.0, 1.0 };
GLfloat ltblue[] = { 0.3, 0.3, 1.0, 1.0 };
GLfloat yellow[] = { 1.0, 1.0, 0.0, 1.0 };
GLfloat cyan[] = { 0.0, 1.0, 1.0, 1.0 };
GLfloat magenta[] = { 0.0, 1.0, 1.0, 1.0 };
GLfloat ltred[] = { 1.0, 0.3, 0.3, 1.0 };
GLfloat central_spotlight[] = { 0.8, 0.9, 1.0, 1.0 };

GLfloat dark_ambient[] = { 0.2, 0.15, 0.1, 1.0 };
GLfloat light_ambient[] = { 0.1, 0.1, 0.1, 1.0 };
GLfloat light_diffuse_twilght[] = { 0.35, 0.2, 0.2, 1.0 };
GLfloat light_diffuse[] = { 0.7, 0.6, 0.5, 1.0 };
GLfloat light_specular[] = { 0.1, 0.1, 0.1, 1.0 };
GLfloat global_ambient[] = { 0.2, 0.2, 0.2, 1.0 };

GLfloat light0_position[] = { -0.62187, 1, 0.78312, 0.0 };

GLfloat light1_position[] = { 25, 2.0+1.9, 0.0+49, 1.0 };
GLfloat spot1_direction[] = { 1.0, -1.0, 0.0 };

GLfloat light2_position[] = { 75, 2.0+1.9, 0.0+49, 1.0 };
GLfloat spot2_direction[] = { -1.0, -1.0, 0.0 };

GLfloat light3_position[] = { 50, 3.9, 25, 1.0 };
GLfloat spot3_direction[] = { 0.0, 0.0, 1.0 };

GLfloat light4_position[] = { 50, -10.0, 50, 1.0 };
GLfloat spot4_direction[] = { 0.0, 1.0, 0.0 };

bool Lighting::lights[4] = {false,false,false,false};

Lighting::Lighting() {
}

Lighting::~Lighting() {
}

void Lighting::Setup() {
  glLightfv(GL_LIGHT1, GL_AMBIENT, black);
  glLightfv(GL_LIGHT1, GL_DIFFUSE, yellow);
  glLightfv(GL_LIGHT1, GL_SPECULAR, yellow);
  glLightf(GL_LIGHT1, GL_SPOT_CUTOFF, 45.0);
  glLightf(GL_LIGHT1, GL_SPOT_EXPONENT, 2.0);

  glLightfv (GL_LIGHT2, GL_AMBIENT, black);
  glLightfv (GL_LIGHT2, GL_DIFFUSE, cyan);
  glLightfv (GL_LIGHT2, GL_SPECULAR, cyan);
  glLightf(GL_LIGHT2, GL_SPOT_CUTOFF, 45.0);
  glLightf(GL_LIGHT2, GL_SPOT_EXPONENT, 2.0);

  glLightfv (GL_LIGHT3, GL_AMBIENT, black);
  glLightfv (GL_LIGHT3, GL_DIFFUSE, ltred);
  glLightfv (GL_LIGHT3, GL_SPECULAR, ltred);
  glLightf(GL_LIGHT3, GL_SPOT_CUTOFF, 45.0);
  glLightf(GL_LIGHT3, GL_SPOT_EXPONENT, 2.0);

  glLightfv (GL_LIGHT4, GL_AMBIENT, black);
  glLightfv (GL_LIGHT4, GL_DIFFUSE, central_spotlight);
  glLightfv (GL_LIGHT4, GL_SPECULAR, central_spotlight);
  glLightf(GL_LIGHT4, GL_SPOT_CUTOFF, 60.0);
  glLightf(GL_LIGHT4, GL_SPOT_EXPONENT, 1.0);

}

void Lighting::Do(std::string timeOfDay) {
  glShadeModel(GL_SMOOTH);
  glColorMaterial(GL_FRONT_AND_BACK, GL_AMBIENT_AND_DIFFUSE);
  glEnable( GL_LIGHTING );
  glEnable( GL_COLOR_MATERIAL );
  glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);

  // light 0 acts as the sun and provides ambient lighting
  if(timeOfDay == "day") {
    light0_position[1] = 1.0;
    glLightModelfv (GL_LIGHT_MODEL_AMBIENT, global_ambient);
    glLightfv (GL_LIGHT0, GL_AMBIENT, dark_ambient);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse);
    glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
  } else if(timeOfDay == "twilight") {
    light0_position[1] = 0.1;
    glLightModelfv (GL_LIGHT_MODEL_AMBIENT, global_ambient);
    glLightfv (GL_LIGHT0, GL_AMBIENT, light_ambient);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, light_diffuse_twilght);
    glLightfv (GL_LIGHT0, GL_SPECULAR, light_specular);
  } else if(timeOfDay == "night") {
    glLightModelfv (GL_LIGHT_MODEL_AMBIENT, dark_ambient);
    glLightfv (GL_LIGHT0, GL_AMBIENT, black);
    glLightfv (GL_LIGHT0, GL_DIFFUSE, black);
    glLightfv (GL_LIGHT0, GL_SPECULAR, black);
  }
  glEnable( GL_LIGHT0 );
  glLightfv (GL_LIGHT0, GL_POSITION, light0_position);

  if(lights[0]) {
    glLightfv (GL_LIGHT1, GL_POSITION, light1_position);
    glLightfv (GL_LIGHT1, GL_SPOT_DIRECTION, spot1_direction);
    glEnable(GL_LIGHT1);
  } else
    glDisable(GL_LIGHT1);

  if(lights[1]) {
    glLightfv (GL_LIGHT2, GL_POSITION, light2_position);
    glLightfv (GL_LIGHT2, GL_SPOT_DIRECTION, spot2_direction);
    glEnable( GL_LIGHT2 );
  } else
    glDisable(GL_LIGHT2);

  if(lights[2]) {
    glLightfv (GL_LIGHT3, GL_POSITION, light3_position);
    glLightfv (GL_LIGHT3, GL_SPOT_DIRECTION, spot3_direction);
    glEnable( GL_LIGHT3 );
  } else
    glDisable(GL_LIGHT3);

  if(lights[3]) {
    glLightfv (GL_LIGHT4, GL_POSITION, light4_position);
    glLightfv (GL_LIGHT4, GL_SPOT_DIRECTION, spot4_direction);
    glEnable( GL_LIGHT4 );
  } else
    glDisable(GL_LIGHT4);

}