#include "Mesh.h"
#include "Lighting.h"

namespace MeshTools {

Mesh::Mesh() {
  numFaces = 0;
  numVertices = 0;
};

void Mesh::Render() {
  if(texMesh == 0) {
    texMesh = SOIL_load_OGL_texture ( "stone.png",//"img_test.png",//
									  SOIL_LOAD_AUTO,
									  SOIL_CREATE_NEW_ID,
									  SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
                    );
  }

  glBindTexture(GL_TEXTURE_2D, texMesh);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

  glEnable(GL_POLYGON_OFFSET_FILL);
  glPolygonOffset( 1.f, 1.f );



  map<string, Group>::iterator iter;
  glEnable(GL_TEXTURE_2D);

  //glDisable(GL_TEXTURE_2D);
  //glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  for(iter = groups.begin(); iter != groups.end(); iter++) {
    Group group = iter->second;

    //Material material = materials[group.material];
    //GLfloat* mat_ambient = material.GetAmbient(); 
    //GLfloat* mat_diffuse = material.GetDiffuse();
    //GLfloat* mat_specular = material.GetSpecular();

    ////glEnable(GL_COLOR_MATERIAL);
    //glMaterialfv(GL_FRONT, GL_DIFFUSE, mat_diffuse);
    //glMaterialfv(GL_FRONT, GL_AMBIENT, mat_ambient);
    //glMaterialfv(GL_FRONT, GL_SPECULAR, mat_specular);
    //glMaterialf(GL_FRONT, GL_SHININESS, material.shininess);
    //glMaterialf(GL_FRONT, GL_EMISSION, material.emission);
    //glLightfv(GL_LIGHT0, GL_AMBIENT, mat_diffuse);

    glEnable(GL_NORMALIZE);
    glPolygonMode(GL_FRONT, GL_FILL);
    glColor3f(1,1,1);
    glBegin(GL_TRIANGLES);
      for(size_t k = 0; k < group.tris.size(); k++) {
        size_t tri_i = group.tris[k];
        if(shadeStyle == ShadeStyle::flat) {
          if(flatNormals.size() == 0)
            SetFlatNormals();
          glNormal3f(flatNormals[tri_i].x, flatNormals[tri_i].y, flatNormals[tri_i].z);
        }

        for(size_t j = 0; j < 3; j++) {
          //glTexCoord2f(j/3 ,(3-j)/3);
          if(k & 1) {
            if(j==0)
              glTexCoord2f(1 ,1);
            else if(j==1)
              glTexCoord2f(0 ,1);
            else if(j==2)
              glTexCoord2f(1 ,0);
          } else {
            if(j==0)
              glTexCoord2f(1 ,1);
            else if(j==1)
              glTexCoord2f(1 ,0);
            else if(j==2)
              glTexCoord2f(0 ,0);
          }

          size_t t = tris[tri_i].vertexIdx[j];
          if(shadeStyle == ShadeStyle::smooth) {
            if(smoothNormals.size() == 0)
              SetSmoothNormals();
            glNormal3f(smoothNormals[t].x, smoothNormals[t].y, smoothNormals[t].z);
          }
          else if(shadeStyle == ShadeStyle::objOriginal) {
            size_t u = tris[tri_i].normalIdx[j];
            glNormal3f(normals[u].x, normals[u].y, normals[u].z);
          }
          //glNormal3f(normals[t].x, normals[t].y, normals[t].z);
          glVertex3f(vertices[t].x, vertices[t].y, vertices[t].z);
        }
      }
    glEnd();

    //// draw normals
    //glDisable(GL_LIGHTING);
    //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    //glBegin(GL_LINES);
    //for(size_t k = 0; k < group.tris.size(); k++) {
    //  size_t tri_i = group.tris[k];
    //  for(size_t j = 0; j < 3; j++) {
    //    size_t t = tris[tri_i].vertexIdx[j];
    //    size_t u = tris[tri_i].normalIdx[j];
    //    glColor3f(1,1,0);
    //    glVertex3f(vertices[t].x, vertices[t].y, vertices[t].z);
    //    glVertex3f(vertices[t].x + smoothNormals[t].x, vertices[t].y + smoothNormals[t].y, vertices[t].z + smoothNormals[t].z);

    //    glColor3f(0,1,1);
    //    glVertex3f(vertices[t].x, vertices[t].y, vertices[t].z);
    //    glVertex3f(vertices[t].x + flatNormals[tri_i].x, vertices[t].y + flatNormals[tri_i].y, vertices[t].z + flatNormals[tri_i].z);

    //    glColor3f(0,0,1);
    //    glVertex3f(vertices[t].x, vertices[t].y, vertices[t].z);
    //    glVertex3f(vertices[t].x + normals[u].x, vertices[t].y + normals[u].y, vertices[t].z + normals[u].z);
    //  }
    //}
    //glEnd();
    //    glColor3f(1,1,1);







    //// draw normals
    //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    //glColor3fv(white);
    //glBegin(GL_LINES);
    //for(size_t k = 0; k < group.tris.size(); k++) {
    //  size_t tri_i = group.tris[k];
    //  for(size_t j = 0; j < 3; j++) {
    //    size_t t = tris[tri_i].vertexIdx[j];
    //    glVertex3f(vertices[t].x, vertices[t].y, vertices[t].z);
    //    //glVertex3f(normals[t].x * 10, normals[t].y * 10, normals[t].z * 10);
    //    glVertex3f(vertices[t].x + smoothNormals[t].x, vertices[t].y + smoothNormals[t].y, vertices[t].z + smoothNormals[t].z);
    //  }
    //}
    //glEnd();

    // triangles
    //glPolygonMode(GL_FRONT_AND_BACK, GL_LINE);
    //glColor3fv(white);
    //glBegin(GL_TRIANGLES);
    //for(size_t k = 0; k < group.tris.size(); k++) {
    //  size_t tri_i = group.tris[k];
    //  for(size_t j = 0; j < 3; j++) {
    //    size_t t = tris[tri_i].vertexIdx[j];
    //    glVertex3f(vertices[t].x, vertices[t].y, vertices[t].z);
    //  }
    //}
    //glEnd();
  }
    //glColor3fv(white);
};

void Mesh::Load(string sFileName) {
// todo -- should mesh or model or MeshBuilder have load?
  //FileReader::Load(sFileName);
};

void Mesh::AddVertex(const Vector3d& vertex) {
  this->vertices.push_back(vertex);
  this->numVertices++;
};

void Mesh::AddNormal(const Vector3d& normal) {
  this->normals.push_back(normal);
  this->numNormals++;
};

void Mesh::AddTexture(const TexCoord& texture) {
  this->texCoords.push_back(texture);
  this->numTexCoords++;
};

inline void Mesh::AddVertexIndex(const int& idx) {
// insert vertex indices to last face
  AddVertexIndex(idx, this->numFaces-1);
}

inline void Mesh::AddNormalIndex(const int& idx) {
// insert texture indices to last face
  AddNormalIndex(idx, this->numFaces-1);
}

inline void Mesh::AddTextureIndex(const int& idx) {
// insert texture indices to last face
  AddTextureIndex(idx, this->numFaces-1);
}

inline void Mesh::AddVertexIndex(const int& idx, const int& faceIdx) {
  assert(faces.size() > 0);
  assert(faces.size() > faceIdx);

  if(idx < 0) {
    faces[faceIdx].vertexIdx.push_back(numVertices + idx);
  } else {
    faces[faceIdx].vertexIdx.push_back(idx);
  }
  ++faces[faceIdx].numVertices;
}

inline void Mesh::AddNormalIndex(const int& idx, const int& faceIdx) {
  if(idx < 0) {
    faces[faceIdx].smoothNormalIdx.push_back(numNormals + idx);
  } else {
    faces[faceIdx].smoothNormalIdx.push_back(idx);
  }
}

inline void Mesh::AddTextureIndex(const int& idx, const int& faceIdx) {
  if(idx < 0) {
    faces[faceIdx].textureIdx.push_back(numTexCoords + idx);
  } else {
    faces[faceIdx].textureIdx.push_back(idx);
  }
}

void Mesh::AddFace(Face& face, const vector<string>& groups /*= defaultGroups*/) {
  faces.push_back(face);
  for(size_t i = 0; i<groups.size()>0;++i) {
    this->groups[groups[i]].faces.push_back(numFaces);
  }
  ++numFaces;

  switch (face.numVertices) {
    case 3 :
      tris.push_back(face);
      for(size_t i = 0; i<groups.size()>0;++i) {
        this->groups[groups[i]].tris.push_back(numTris);
      }
      ++numTris;
      break;
    case 4 :
      quads.push_back(face);
      for(size_t i = 0; i<groups.size()>0;++i) {
        this->groups[groups[i]].quads.push_back(numQuads);
      }
      ++numQuads;
      break;
    default :
      polys.push_back(face);
      for(size_t i = 0; i<groups.size()>0;++i) {
        this->groups[groups[i]].polys.push_back(numPolys);
      }
      ++numPolys;
      break;
  }

};
////////////////////////////////////////////////////////

void Mesh::Clear() {
  vertices.clear();

  faces.clear();
  quads.clear();
  tris.clear();
  polys.clear();

  smoothNormals.clear();
  normals.clear();
  flatNormals.clear();
  texCoords.clear();

  materials.clear();
  groups.clear();

  numVertices = 0;
  numNormals = 0;
  numFaces = 0;
  numTris = 0;
  numQuads = 0;
  numPolys = 0;
  numTexCoords = 0;
}

void Mesh::SetNormals() {
  // set normal = flatNormals
  if(flatNormals.size()==0)
    SetFlatNormals();

  normals.assign(flatNormals.begin(), flatNormals.end());

  for(size_t i = 0; i<faces.size(); ++i) {
    faces[i].smoothNormalIdx.assign(faces[i].flatNormalIdx.begin(), faces[i].flatNormalIdx.end());
  }
  for(size_t i = 0; i<tris.size(); ++i) {
    tris[i].smoothNormalIdx.assign(tris[i].flatNormalIdx.begin(), tris[i].flatNormalIdx.end());
  }
  for(size_t i = 0; i<quads.size(); ++i) {
    quads[i].smoothNormalIdx.assign(quads[i].flatNormalIdx.begin(), quads[i].flatNormalIdx.end());
  }
  for(size_t i = 0; i<polys.size(); ++i) {
    polys[i].smoothNormalIdx.assign(polys[i].flatNormalIdx.begin(), polys[i].flatNormalIdx.end());
  }
}

void Mesh::SetFlatNormals(vector<Face> &faces, vector<Vector3d>& flatNormals) {
  // determine the normal for each face and assign to each face vertex
  for(size_t i = 0; i < faces.size(); i++) {
    if(faces[i].numVertices > 2) {
      int iV1 = faces[i].vertexIdx[0];
      int iV2 = faces[i].vertexIdx[1];
      int iV3 = faces[i].vertexIdx[2];

      // normal should point AWAY from the centre
      Vector3d flatNormal = Normalize( CrossProduct((vertices[iV1] - vertices[iV2]), (vertices[iV1] - vertices[iV3])));
      //Vector3d flatNormal = Normalize( CrossProduct((vertices[iV1] - vertices[iV3]), (vertices[iV1] - vertices[iV2])));
      faces[i].flatNormalIdx.push_back(flatNormals.size());
      flatNormals.push_back(flatNormal);
    }
  }
}

void Mesh::SetFlatNormals() {
  flatNormals.clear();
  for(size_t i = 0; i<faces.size(); ++i) {
    faces[i].flatNormalIdx.clear();
  }
  for(size_t i = 0; i<tris.size(); ++i) {
    tris[i].flatNormalIdx.clear();
  }
  for(size_t i = 0; i<quads.size(); ++i) {
    quads[i].flatNormalIdx.clear();
  }
  for(size_t i = 0; i<polys.size(); ++i) {
    polys[i].flatNormalIdx.clear();
  }
  numNormals=0;
  SetFlatNormals(tris, flatNormals);
  SetFlatNormals(quads, flatNormals);
  SetFlatNormals(polys, flatNormals);
}

void Mesh::ScaleBy(const float factor) {
	for(size_t i = 0; i < vertices.size(); i++)
		vertices[i] *= factor;
}

void Mesh::SetSmoothNormals() {
  if(flatNormals.size()==0)
    SetFlatNormals();

  // for each vertex build a list of all faces it belongs to
  // quick way of doing this
  //   set up vector for all vertices
  //   for each face note which vertices it contains

  vector<vector<int> > vertexBelongsToFaces(numVertices);
  for(size_t i = 0; i < tris.size(); i++) {
    for(size_t j = 0; j < tris[i].numVertices; j++) {
      int vertexIdx = tris[i].vertexIdx[j];
      int data = tris[i].flatNormalIdx[0];
      vertexBelongsToFaces[vertexIdx].push_back(data);
    }
  }
  for(size_t i = 0; i < quads.size(); i++) {
    for(size_t j = 0; j < quads[i].numVertices; j++) {
      int vertexIdx = quads[i].vertexIdx[j];
      int data = quads[i].flatNormalIdx[0];
      vertexBelongsToFaces[vertexIdx].push_back(data);
    }
  }
  for(size_t i = 0; i < polys.size(); i++) {
    for(size_t j = 0; j < polys[i].numVertices; j++) {
      int vertexIdx = polys[i].vertexIdx[j];
      int data = polys[i].flatNormalIdx[0];
      vertexBelongsToFaces[vertexIdx].push_back(data);
    }
  }

  // for each face count and sum unique normals
  for(size_t i = 0; i < vertexBelongsToFaces.size(); ++i) {
    Vector3d t = Vector3d(0,0,0);
    size_t end = vertexBelongsToFaces[i].size();
    for(size_t j = 0; j < end; ++j) {
      size_t k = j+1;
      size_t idxFace1 = vertexBelongsToFaces[i][j];

      while (k < end) {
        int idxFace2 = vertexBelongsToFaces[i][k];

        if(flatNormals[idxFace1] == flatNormals[idxFace2]) {
          break;
        } else {
          ++k;
        }
      }
      if(k==end) {
        t += flatNormals[idxFace1];
      }
    }
    Vector3d t1 = Normalize(t);

    smoothNormals.push_back(t1);
  }
}

void Mesh::FindBoundingBox() {
  if  (vertices.size() == 0)
    return;
  upperBound = vertices[0];
  lowerBound = vertices[0];

  for(size_t i = 1, size = vertices.size(); i < size; i++) {
    for(int j = 0; j<3; j++) {
      if(upperBound[j] < vertices[i][j])
        upperBound[j] = vertices[i][j];
      if(lowerBound[j] > vertices[i][j])
        lowerBound[j] = vertices[i][j];
    }
  }

  center = (upperBound + lowerBound) /2;
}

void Mesh::InvertNormals() {
  for(size_t i = 0, size = normals.size(); i < size; i++) {
    normals[i] = -normals[i];
  }
  for(size_t i = 0, size = smoothNormals.size(); i < size; i++) {
    smoothNormals[i] = -smoothNormals[i];
  }
  for(size_t i = 0, size = flatNormals.size(); i < size; i++) {
    flatNormals[i] = -flatNormals[i];
  }

}

void Mesh::Centralise() {
  for(size_t i = 0, size = vertices.size(); i < size; i++) {
    vertices[i] = vertices[i] - center;
  }

  center = Vector3d(0,0,0);
}

void Mesh::SetMaterial(vector<string> groups, string material) {
  for(size_t i = 0; i < groups.size(); ++i) {
    this->groups[groups[i]].material = material;
  }
};

void Mesh::AddMaterial(string name, Material material) {
  this->materials[name] = material;
};

void Mesh::AddTextureName(string name, string pathname) {
  this->textures[name].fullpath = pathname;
  this->textures[name].name = name;
};

/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
// helper functions
// non-member non-friend functions 
//  cf Scott Meyers, How Non-Member Functions Improve Encapsulation 
//  http://www.ddj.com/cpp/184401197

void AddVertices(Mesh mesh, const vector<Vector3d> vertices) {
  for(size_t i=0; i<vertices.size(); ++i) {
    mesh.AddVertex(vertices[i]);
  }
};

void AddNormals(Mesh mesh, const vector<Vector3d> normals) {
  for(size_t i=0; i<normals.size(); ++i) {
    mesh.AddNormal(normals[i]);
  }
};

void AddVertexIndices(Mesh mesh, const vector<int> vertexIndices) {
  for(size_t i=0; i<vertexIndices.size(); ++i) {
    mesh.AddVertexIndex(vertexIndices[i]);
  }
};

void AddTextures(Mesh mesh, const vector<TexCoord> textures){
  for(size_t i=0; i<textures.size(); i+=2) {
    mesh.AddTexture(textures[i]);
  }
};

void AddNormalIndices(Mesh mesh, const vector<int> normalIndices) {
  for(size_t i=0; i<normalIndices.size(); i++) {
    mesh.AddNormalIndex(normalIndices[i]);
  }
};

void AddTextureIndices(Mesh mesh, const vector<int> textureIndices) {
  for(size_t i=0; i<textureIndices.size(); i++) {
    mesh.AddTextureIndex(textureIndices[i]);
  }
};

inline void AddVertexIndex(Mesh mesh, const string idx, const int faceIdx) {
  mesh.AddVertexIndex(val_from_string<int>(idx, std::dec), faceIdx);
}

inline void AddNormalIndex(Mesh mesh, const string idx, const int faceIdx) {
  mesh.AddNormalIndex(val_from_string<int>(idx, std::dec), faceIdx);
}

inline void AddTextureIndex(Mesh mesh, const string idx, const int faceIdx) {
  mesh.AddTextureIndex(val_from_string<int>(idx, std::dec), faceIdx);
}

}