#include "IFileReader.h"
#include "ObjFileReader.h"

PtrToFunStream ObjFileReader::pfFaceStream = NULL;
PtrToFun ObjFileReader::pfFace = NULL;

ifstream m_fileStream;

const string defaultGroupname = "default";

map<FunctionIDs, PtrToFunVecStr> ObjFileReader::pfExternal_MapFuncIdToFunc;
bool ObjFileReader::fnPtrsAreSet = false;

////////////////////////////////////////////////////
bool ObjFileReader::FnPtrsAreSet() {
  return fnPtrsAreSet;
}

bool ObjFileReader::CanRead(string filename) {
  size_t lastPos = filename.find_last_of(".obj");

  return (lastPos == filename.size() - 1);
};

void ObjFileReader::LoadMTL(string sFileName) {
  vector<string> material;
  ifstream fileStream(sFileName.c_str());
  if (fileStream.good()) {
    while (!fileStream.eof()) {
      string keyword, r, g, b;
      fileStream >> keyword;

      switch(keyword[0]) {
        case 'n' : {
          // newmtl - if there's a material pending add it          
          if(material.size()>0)
            (*pfExternal_MapFuncIdToFunc[addMaterial])(material);
          material.clear();

          fileStream >> keyword;
          material.push_back(keyword);
          break; 
          }
        case 'K' : {
          // Ka/Kd
          fileStream >> r >> g >> b;
          material.push_back(keyword);
          material.push_back(r);
          material.push_back(g);
          material.push_back(b);
          break;
          }
        case 'N' :  // Ns
        case 'i' :  {
          // illum
          fileStream >> r;
          material.push_back(keyword);
          material.push_back(r);
          break;
          }
        case 'm' : {
          // map_Ka/map_Kd
          material.push_back(keyword);
          string line;
          getline(fileStream, line);
          istringstream iss(line);
          while (iss >> r)
          {
            // for now ignore fields except the last = filename
            g = r;
          }
          material.push_back(g);
          break;
          }
		//case 'T' : { } // Tr
        default :
          getline(fileStream, keyword); // key word not recognised, swallow the remainder of the line
      }
    }
  } else {
    cerr << "couldn't open material file: " << sFileName << endl;
  }
  fileStream.close();

  // if there's a material pending add it
  if(material.size()>0)
    (*pfExternal_MapFuncIdToFunc[addMaterial])(material);
};

string streamPath(const string& fullpath) {
  string path("");
  size_t pos = fullpath.find_last_of("\\");
  if(pos==string::npos)
    pos = fullpath.find_last_of("/");
  if(pos!=string::npos)
    path = fullpath.substr(0,++pos);

  return path;
}

void ObjFileReader::Load(string sFileName) {
  // if pointers haven't been set do so now
  // external function pointers are set externally
  if(fnPtrsAreSet == false) {
    for(FunctionIDs i = FunctionIDs_start; i < FunctionIDs_end; ++i) {
      map<FunctionIDs, PtrToFunVecStr>::iterator iter = pfExternal_MapFuncIdToFunc.find(i);
      if(iter == pfExternal_MapFuncIdToFunc.end())
        pfExternal_MapFuncIdToFunc[i] = ObjFileReader::NOP;
    }
    fnPtrsAreSet = true;
  }
  pfFace = ObjFileReader::FindFaceLayout;
  
  m_fileStream.clear();
  m_fileStream.open(sFileName.c_str());
  if (m_fileStream.good()) {
    while (!m_fileStream.eof()) {
      string keyword;
      m_fileStream >> keyword;
      
      if (m_fileStream.fail())
        break;

      switch(keyword[0]) {
        case 'v' : {
          switch(keyword[1]) {
            case 't' : Texture(); break;
            case 'n' : Normal();  break;
            default  : Vertex();
          }
          };break;
        case 'f' : (*pfFace)();   break;
        case 'g' : Group();       break;
        case 'u' : UseMaterial(); break;
        case 'm' : MtlLib(streamPath(sFileName));      break;
        default  : getline(m_fileStream, keyword); // key word not recognised, swallow the remainder of the line         
      }
    }
  } else {
    cerr << "couldn't open object file:" << sFileName << endl;
  }
  m_fileStream.close();
}

void ObjFileReader::FaceV(const string& line){
  vector<string> vertexIndices; 
  size_t pos0 = 0;
  size_t pos1 = 0;

  while(true) {
    pos1 = line.find_first_of(" ", pos0);
    vertexIndices.push_back(line.substr(pos0, (pos1 - pos0)));

    if(pos1 == string::npos)
      break;

    pos0 = line.find_first_not_of(" ", pos1+1);
    if(pos0 == string::npos)
      break;
  }

  (*pfExternal_MapFuncIdToFunc[addFaceV])(vertexIndices);
};

void ObjFileReader::FaceVN(const string& line){
  vector<string> indices; 
  size_t pos0 = 0;
  size_t pos1 = 0;

  while(true) {
    pos1 = line.find_first_of("//", pos0);  //find vertex
  
    if(pos1 == string::npos)
      break;

    indices.push_back(line.substr(pos0, (pos1 - pos0)));
  
    ++pos1; ++pos1;
    pos0 = line.find_first_of(" ", pos1);  // find normal
    indices.push_back(line.substr(pos1, (pos0 - pos1)));

    if(pos0 == string::npos)
      break;
  }

  (*pfExternal_MapFuncIdToFunc[addFaceVN])(indices);
};

void ObjFileReader::FaceVT(const string& line){
  vector<string> indices; 
  size_t pos0 = 0;
  size_t pos1 = 0;

  while(true) {
    pos1 = line.find_first_of("/", pos0); //find vertex

    if(pos1 == string::npos)
      break;

    indices.push_back(line.substr(pos0, (pos1 - pos0)));

    ++pos1;
    pos0 = line.find_first_of(" ", pos1); // find normal
    indices.push_back(line.substr(pos1, (pos0 - pos1)));

    if(pos0 == string::npos)
      break;
  }

  (*pfExternal_MapFuncIdToFunc[addFaceVT])(indices);
};

void ObjFileReader::FaceVTN(const string& line){
  vector<string> indices; 
  size_t pos0 = 0;
  size_t pos1 = 0;

  while(true) {
    pos1 = line.find_first_of("/", pos0); //find vertex

    if(pos1 == string::npos)
      break;

    indices.push_back(line.substr(pos0, (pos1 - pos0)));
   
    ++pos1;
    pos0 = line.find_first_of("/", pos1); // find texture
    indices.push_back(line.substr(pos1, (pos0 - pos1)));

    ++pos0;
    pos1 = line.find_first_of(" ", pos0); // find normal
    indices.push_back(line.substr(pos0, (pos1 - pos0)));

    if(pos1 == string::npos)
      break;

    pos0 = pos1;
  }

  (*pfExternal_MapFuncIdToFunc[addFaceVTN])(indices);
};

void ObjFileReader::SetExternalFunctionPointers(map<FunctionIDs, PtrToFunVecStr> pfExternal) {
	pfExternal_MapFuncIdToFunc = pfExternal;
  fnPtrsAreSet = true;
}

void ObjFileReader::Vertex(){
  vector<string> vector;
  string f1, f2, f3;

  m_fileStream >> f1 >> f2 >> f3;
  vector.push_back(f1);
  vector.push_back(f2);
  vector.push_back(f3);

  (*pfExternal_MapFuncIdToFunc[addVertex])(vector);
};

void ObjFileReader::Normal() {
  string keyword, f0, f1, f2, f3;
  m_fileStream >> f1 >> f2 >> f3;

  vector<string> normal;
  normal.push_back(f1);
  normal.push_back(f2);
  normal.push_back(f3);

  (*pfExternal_MapFuncIdToFunc[addNormal])(normal);
};

void ObjFileReader::Texture() {
  string f1, f2;
  m_fileStream >> f1 >> f2;

  vector<string> textureCoord;
  textureCoord.push_back(f1);
  textureCoord.push_back(f2);

  (*pfExternal_MapFuncIdToFunc[addTexture])(textureCoord);
};

void ObjFileReader::FaceV() {
  string f1;
  vector<string> vertexIndices; 

  m_fileStream >> f1;
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1);

  string line;
  getline(m_fileStream, line);
  istringstream iss(line);
  while (iss >> f1)
  {
    vertexIndices.push_back(f1);
  }

  (*pfExternal_MapFuncIdToFunc[addFaceV])(vertexIndices);
};

void ObjFileReader::FaceVT() {
  string f1;
  vector<string> vertexIndices;

  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1);

  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1);

  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1);

  string line;
  getline(m_fileStream, line);
  istringstream iss(line);
  while (iss >> f1)
  {
    size_t pos = line.find("/");
    f1 = line.substr(0,pos);
    vertexIndices.push_back(f1);
    f1 = line.substr(pos+1);
    vertexIndices.push_back(f1);
  }

  (*pfExternal_MapFuncIdToFunc[addFaceVT])(vertexIndices);
};

void ObjFileReader::FaceVN() {
  string f1;
  vector<string> vertexIndices;

  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1.substr(1));

  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1.substr(1));

  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1.substr(1));

  string line;
  getline(m_fileStream, line);
  istringstream iss(line);
  while (iss >> f1)
  {
    size_t pos = line.find("/");
    f1 = line.substr(0,pos);
    vertexIndices.push_back(f1);
    f1 = line.substr(pos+2);
    vertexIndices.push_back(f1);
  }

  (*pfExternal_MapFuncIdToFunc[addFaceVN])(vertexIndices);
};

void ObjFileReader::FaceVTN() {
  string f1;
  vector<string> vertexIndices;

  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1);

  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1);

  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  getline(m_fileStream, f1, '/');
  vertexIndices.push_back(f1);
  m_fileStream >> f1;
  vertexIndices.push_back(f1);

  string line;
  getline(m_fileStream, line);
  istringstream iss(line);
  while (iss >> f1)
  {
    size_t pos = line.find("/");
    f1 = line.substr(0,pos);
    vertexIndices.push_back(f1);
    size_t pos1 = line.find("/", pos+1);
    f1 = line.substr(pos+1, pos1-pos-1);
    vertexIndices.push_back(f1);
    f1 = line.substr(pos1+1);
    vertexIndices.push_back(f1);
  }

  (*pfExternal_MapFuncIdToFunc[addFaceVTN])(vertexIndices);
};

void ObjFileReader::FindFaceLayout() {
  // determine whether file contains
  // f v
  // f v/n
  // f v/n/t
  // f v//t
  // should be corensistent for entire file
  string f1;
  m_fileStream >> f1;

  string line = "";
  getline(m_fileStream, line);
  line = f1 + " " + line;

  int n1 = f1.find("/");
  if(n1 == string::npos){
    pfFace = FaceV;
    FaceV(line);
  } else {
    int n2 = f1.find("/",n1+1);
    if(n2 == string::npos){
      pfFace = FaceVT;
      FaceVT(line);
    } else {
      if(n2-n1 == 1) {
        pfFace = FaceVN;
        FaceVN(line);
      } else {
        pfFace = FaceVTN;
        FaceVTN(line);
      }
    }
  }
};

void ObjFileReader::Group() {
  string line, f1;
  vector<string> groups;

  getline(m_fileStream, line);
  istringstream iss(line);
  while (iss >> f1)
  {
    groups.push_back(f1);
  }

  (*pfExternal_MapFuncIdToFunc[setGroups])(groups);
};

void ObjFileReader::UseMaterial() {
  string mtl;
  m_fileStream >> mtl;
  vector<string> material;
  material.push_back(mtl);
  (*pfExternal_MapFuncIdToFunc[setMaterial])(material);
};


void ObjFileReader::MtlLib(string& path) {
  string line, mtlLibName;

  getline(m_fileStream, line);
  istringstream iss(line);
  while (iss >> mtlLibName)
  {
    LoadMTL(path + mtlLibName);
  }
};
