#ifndef _OBJ_FILE_READER_H
#define _OBJ_FILE_READER_H

#include <string>
#include <fstream>
#include <vector>
#include <map>
#include <algorithm>
#include <iostream>
#include <sstream> 
#include "vector_inl.h"
#include "IFileReader.h"

typedef void (*PtrToFunVecStr)(const std::vector<std::string>&);
typedef void (*PtrToFunStream)(std::ifstream&);
typedef void (*PtrToFun)();

using namespace std;

class ObjFileReader {
private:
  static bool fnPtrsAreSet;
  // pointers to external functions
  static map<FunctionIDs, PtrToFunVecStr> pfExternal_MapFuncIdToFunc;

  // dummy no op function
  static void NOP(const std::vector<std::string>& ) { };

public:
// methods
  static void Load(std::string sFileName);
  static void LoadMTL(std::string sFileName);
  static bool CanRead(std::string fileName);
  static bool FnPtrsAreSet();

  //parse data from std::string
  static void FaceV(const std::string& line);
  static void FaceVT(const std::string& line);
  static void FaceVN(const std::string& line);
  static void FaceVTN(const std::string& line);

  // parse data from stream
  static void Vertex();
  static void Normal();
  static void Texture();
  static void Group();
  static void UseMaterial();
  // Face isn't _directly_ supported since there are several possible layouts
  static void FaceV();
  static void FaceVT();
  static void FaceVN();
  static void FaceVTN();
  // but first we have identify which one to use
  static void FindFaceLayout();

  static void MtlLib(std::string& path);


  static void SetExternalFunctionPointers(map<FunctionIDs, PtrToFunVecStr> pfExternal);

// data
  // pointer to internal function
  static PtrToFunStream pfFaceStream;
  static PtrToFun pfFace;
  //ifstream m_fileStream;
};

#endif