#include "SkyMap.h"

const GLfloat X1 = 100, X2 = -100;
const GLfloat Y1 = 100, Y2 = -100;
const GLfloat Z1 = -100, Z2 = 100;
const GLfloat data_cube[] = {
  X1, Y1, Z1, // 0
  X2, Y1, Z1, // 1
  X2, Y1, Z2, // 2
  X1, Y1, Z2, // 3
  X1, Y2, Z1, // 4
  X2, Y2, Z1, // 5
  X2, Y2, Z2, // 6
  X1, Y2, Z2, // 7

};

// indices
const  GLushort cubeIndices[]= {
  0,1,2,3, //top    0
  4,0,3,7, //left   1
  7,3,2,6, //back   2
  6,2,1,5, //right  3
  5,1,0,4, // 1,5,4,0,////front
  4,7,6,5, //bottom 5
};

void SkyMap::Init() {
  // textures from 
  // www.reallyreallygoodthings.com/starterpack1/StarterPackFullv2.zip
  texSky.resize(10);
  for(unsigned int i = 0; i < 5; i++) {
	  std::string mapID("cloudy/" + val_to_string(i) + ".jpg");
	  texSky[i] = SOIL_load_OGL_texture ( mapID.c_str(),
										SOIL_LOAD_AUTO,
										SOIL_CREATE_NEW_ID,
										SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
	  );
	  mapID = "sunset/" + val_to_string(i) + ".jpg";
	  texSky[i+5] = SOIL_load_OGL_texture ( mapID.c_str(),
										SOIL_LOAD_AUTO,
										SOIL_CREATE_NEW_ID,
										SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
	  );
  }
}

void SkyMap::Render(const Vector3d & center, std::string timeOfDay ) {
  unsigned int textureOffset = 0;
  if(timeOfDay == "night")
    return;
  if(timeOfDay == "twilight")
    textureOffset = 5;

  glPushClientAttrib( GL_CLIENT_VERTEX_ARRAY_BIT );

  glDisable( GL_LIGHTING );
  glEnable(GL_TEXTURE_2D);
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  glDepthMask(GL_FALSE);
  glDisable(GL_DEPTH_TEST);

  glTranslatef(center.x, center.y, center.z);

  for(unsigned int i = 0; i<5; i++) {
    glBindTexture(GL_TEXTURE_2D, texSky[i + textureOffset]);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
    glBegin(GL_QUADS);
    glTexCoord2f(0.0 ,0.0);  
    glVertex3fv(&data_cube[cubeIndices[i * 4 + 0] * 3]);

    glTexCoord2f(0.0 ,1.0);
    glVertex3fv(&data_cube[cubeIndices[i * 4 + 1] * 3]);

    glTexCoord2f(1.0, 1.0);  
    glVertex3fv(&data_cube[cubeIndices[i * 4 + 2] * 3]);

    glTexCoord2f(1.0, 0.0);  
    glVertex3fv(&data_cube[cubeIndices[i * 4 + 3] * 3]);
    glEnd();
  }

  glTranslatef(-center.x, -center.y, -center.z);

  glDisable(GL_VERTEX_ARRAY);
  glEnable(GL_DEPTH_TEST);
  glDepthMask(GL_TRUE);
  glPopClientAttrib(  );
}