#include "Terrain.h"

Terrain::Terrain() {
  InitTerrain(15, 15);
}

Terrain::~Terrain() {
}

Terrain::Terrain(unsigned int rows, unsigned int cols) {
  InitTerrain(rows, cols);
}
void Terrain::InitTexture() {
  texGround = SOIL_load_OGL_texture ( "ground.png",
                    SOIL_LOAD_AUTO,
                    SOIL_CREATE_NEW_ID,
                    SOIL_FLAG_MIPMAPS | SOIL_FLAG_INVERT_Y | SOIL_FLAG_NTSC_SAFE_RGB | SOIL_FLAG_COMPRESS_TO_DXT
                    );

}

void Terrain::InitTerrain(unsigned int rows, unsigned int cols) {
  GLubyte * heightmap;
  int width = 0, height = 0, channels = 0;
  heightmap = SOIL_load_image (
            "heightmap.png",
            &width, &height, &channels,
            SOIL_LOAD_AUTO
  );  

  m_rows = rows;
  m_cols = cols;
  float xstep=.01*m_rows/PI;
  float zstep=.01*m_cols/PI;

  m_vertices1.resize(m_rows * m_cols);

  int c = 0;
  float row_size = 1/float(m_rows);
  float col_size = 1/float(m_cols);
  float tr = 0;
  float tc = 0;
  for(int x = 0; x < m_cols; x++) {
    for(int z = 0; z < m_rows; z++) {
      m_vertices1[x * m_rows + z].x = x;
      m_vertices1[x * m_rows + z].y = 0;
      m_vertices1[x * m_rows + z].z = z;
      // texture
      m_vertices1[x * m_rows + z].s0 = tr;
      m_vertices1[x * m_rows + z].t0 = tc;

	    // temp set normal vertical
      float ht = heightmap[x * m_rows + z];
      ht *= .137;
      ht -= 30.0f;
      m_vertices1[x * m_rows + z].y = ht;

      m_vertices1[x * m_rows + z].nx = 0;
      m_vertices1[x * m_rows + z].ny = 1;
      m_vertices1[x * m_rows + z].nz = 0;

      tr += row_size;
    }
    tc += col_size;
    tr = 0;
  }

  // init vertex list
  m_strips.resize(m_cols-1);
  for(int i = 0; i < m_cols-1; i++) {
    m_strips[i].resize(m_rows * 2);
  }

  for(int i = 0; i < m_cols-1; i++) {
    for(int j = 0; j < m_rows; j++) {
      m_strips[i][j*2] = m_rows * i + j;//0;//
      m_strips[i][j*2+1] = m_rows * (i+1) + j;//58;//
    }
  }

  SOIL_free_image_data( heightmap );
}


void Terrain::Render() {
  glEnable(GL_TEXTURE_2D);
  glEnable( GL_LIGHTING );
  glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

  glBindTexture(GL_TEXTURE_2D, texGround);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  glEnable(GL_TEXTURE_2D);

  glEnableClientState( GL_VERTEX_ARRAY );
  glEnableClientState( GL_NORMAL_ARRAY );
  glEnableClientState( GL_TEXTURE_COORD_ARRAY );

  glVertexPointer(3, GL_FLOAT, sizeof(MyVertex), &m_vertices1[0]);
  glTexCoordPointer(2, GL_FLOAT, sizeof(MyVertex), &m_vertices1[0].s0);
  glNormalPointer(GL_FLOAT, sizeof(MyVertex), &m_vertices1[0].nx);

  for(int i=0; i<m_cols-1; i++) {
    glDrawElements( GL_TRIANGLE_STRIP, m_strips[i].size(), GL_UNSIGNED_INT, &m_strips[i][0] );
  }

  glDisable(GL_TEXTURE_2D);
};

float Terrain::GetHeight(float x, float z) {
  return m_vertices1[int(x) * m_rows + int(z)].y;
};
