#include "Viewer.h"

Viewer::Viewer() {
	position = Vector3d(0,2,0);
	up = Vector3d(0,1,0);
	forward = Vector3d(1,0,0);
	velocity = 0;
  strafeVelocity = 0;
	angularVelocity = 0.0f;
  strafeMaxVelocity = 5.0f;
	maxVelocity = 5.f;
	maxAngularVelocity = 90.0;
	tiltAngle = 0.0f;
}

Viewer::~Viewer() {
}

double Viewer::Tilt() const{
	return tiltAngle;
};

Vector3d Viewer::Up() {
	return up;
};

Vector3d Viewer::LookAt() {
	return forward + position;
};

Vector3d Viewer::Direction(){
	return forward;
};

Vector3d Viewer::Position(){
	return position;
};

void Viewer::SetHeight(double newHeight) {
	height = newHeight;
};

void Viewer::ChangeHeightBy(double newHeight) {
	height += newHeight;
  position.y += newHeight;
};

void Viewer::SetPosition(const Vector3d & newPostion) {
	position = newPostion;
}

const Vector3d & Viewer::GetPosition() {
	return position;
}

const Vector3d & Viewer::GetForward() {
	return forward;
}

const double & Viewer::GetVelocity() {
	return velocity;
}

const double & Viewer::GetAcceleration() {
	return acceleration;
}

void Viewer::Tilt(double degrees){
	tiltAngle+=degrees;
	if(tiltAngle > 80)
		tiltAngle = 80;
	else if(tiltAngle < -80)
		tiltAngle = -80;
};


void Viewer::TiltTo(double degrees){
	tiltAngle=degrees;
};

void Viewer::Strafe(double impulse) {
	strafeAcceleration = impulse;
};

void Viewer::Accelerate(double newAcceleration) {
	acceleration = newAcceleration;
};

void Viewer::Turn(double angle) {
	angularVelocity += angle;
	if(fabs(angularVelocity) > maxAngularVelocity)
		angularVelocity = sgn(angularVelocity)*maxAngularVelocity;
	return;
};

void Viewer::SetLimits(Vector3d upper, Vector3d lower) {
	upperLimit = upper;
	lowerLimit = lower;
}

void Viewer::ClampToLimits() {
	if(position.x > upperLimit.x)
		position.x = upperLimit.x;
	else if(position.x < lowerLimit.x)
		position.x = lowerLimit.x;

	if(position.z > upperLimit.z)
		position.z = upperLimit.z;
	else if(position.z < lowerLimit.z)
		position.z = lowerLimit.z;
}

void Viewer::Update(const double & ms, double groundLevel) {
  Update(ms);
  position.y = height + groundLevel;
}

bool Viewer::IsNotActive() {
  return velocity == 0.0 && acceleration == 0.0 
      && angularVelocity == 0.0 && angularAcceleration == 0.0
      && strafeVelocity == 0.0
      && angularTiltVelocity == 0.0;
};

bool Viewer::IsActive() {
  return IsNotActive() == true;
};

void Viewer::Update(const double & ms) {
	if(ms==0)
    return;
  // if maximum velocity is reached ignore acceleration
  // x = vt
  // v = v0 + at
  // x = vt = v0t + at*t
  // move x = ut + 1/2at*t
  { // walk
    double newVel = velocity + acceleration * ms;
    double xchange;
    double zchange;

    if(abs(newVel) > maxVelocity) {
      if(newVel > maxVelocity)
        newVel = maxVelocity;
      else
        newVel = -maxVelocity;
      double averageVel = (velocity + newVel) / 2;
      xchange = forward.x * averageVel * ms;
      zchange = forward.z * averageVel * ms;
    } else {
      xchange = forward.x * (velocity + 0.5 * acceleration * ms) * ms;
      zchange = forward.z * (velocity + 0.5 * acceleration * ms) * ms;
    }
    velocity = newVel;

    position.x += xchange;
    position.z += zchange;
  }

  {	// strafe
    double newVel = strafeVelocity + strafeAcceleration * ms;
    double xchange;
    double zchange;
	  double tx = -forward.z;
	  double tz = forward.x;

    if(abs(newVel) > strafeMaxVelocity) {
      if(newVel > strafeMaxVelocity)
        newVel = strafeMaxVelocity;
      else
        newVel = -strafeMaxVelocity;
      double averageVel = (strafeVelocity + newVel) / 2;
      xchange = tx * averageVel * ms;
      zchange = tz * averageVel * ms;
    } else {
      xchange = tx * (strafeVelocity + 0.5 * strafeAcceleration * ms) * ms;
      zchange = tz * (strafeVelocity + 0.5 * strafeAcceleration * ms) * ms;
    }
    strafeVelocity = newVel;

    position.x += xchange;
    position.z += zchange;
  }

	// rotate
	double rads = angularVelocity * ms * PI/180;
		
	double tx = cos(rads) * forward.x - sin(rads) * forward.z;
	forward.z = cos(rads) * forward.z + sin(rads) * forward.x;
	forward.x = tx;

	ClampToLimits();

	// dampen velocities
	if(acceleration == 0)
    velocity *= 0.8f;
	angularVelocity *= 0.9f;
	if(strafeAcceleration == 0)
	  strafeVelocity *= 0.8f;

	if(fabs(velocity) < 0.2f)
		velocity = 0.0f;

	if(fabs(angularVelocity) < 0.2f)
		angularVelocity = 0.0f;

	if(fabs(strafeVelocity) < 0.2f)
		strafeVelocity = 0.0f;
}

void Viewer::RotateYby(const double& angle) {
	double rads = angle * PI/180;
		
	double tx = cos(rads) * forward.x - sin(rads) * forward.z;
	forward.z = cos(rads) * forward.z + sin(rads) * forward.x;
	forward.x = tx;
};

void Viewer::RotateYto(const double& angle) {
	double rads = angle * PI/180;
  forward = Vector3d(1,0,0); 
		
	double tx = cos(rads) * forward.x - sin(rads) * forward.z;
	forward.z = cos(rads) * forward.z + sin(rads) * forward.x;
	forward.x = tx;
};

void Viewer::PitchBy(const double& angle) {
	double rads = angle * PI/180;
		
	double tz = cos(rads) * forward.z - sin(rads) * forward.y;
	forward.y = cos(rads) * forward.y + sin(rads) * forward.z;
	forward.z = tz;

	tz = cos(rads) * up.z - sin(rads) * up.y;
	up.y = cos(rads) * up.y + sin(rads) * up.z;
	up.z = tz;
}

