// coursework - Carl Bateman
// Student ID 080045108
// INM376: COMPUTER GRAPHICS - 2009


#include "GLee.h"
#include <stdlib.h>
#include <GL/glut.h>
#include "soil.h"
#include <stdio.h>
#include <vector>
#include "vector_inl.h"
#include "terrain.h"
#include "viewer.h"
#include "skymap.h"
#include <string>
#include "Mesh.h"
#include "MeshDataLoader.h"
#include "Lighting.h"
#include "BannerScreen.h"

void idleFunc();
void startIdleFunc();

//#define development

#define KEYDOWN(vk_code) ((GetAsyncKeyState(vk_code) & 0x8000) ? 1 : 0)
#define KEYUP(vk_code)   ((GetAsyncKeyState(vk_code) & 0x8000) ? 0 : 1)
char keys[257];

GLint windowX=800;
GLint windowY=600;
GLint Xpos=0;
GLint Ypos=0;
bool fullscreen = false;
DWORD timLast;
std::string timeOfDay("day");
Viewer viewer;
Terrain terrain(100,100);
SkyMap skyMap;
Mesh mesh;
BannerScreen banner;

struct CameraSetUp {
  Vector3d position;
  double tilt;
  double turn;
};
CameraSetUp cameraSetups[5];

struct Transition {
	int target;
	Vector3d stepSize;
};
Transition transition;

Vector3d posHenge(50, 1.9, 49);

void display() {
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

  glLoadIdentity();

  glRotatef(viewer.Tilt(), 1.0, 0.0, 0.0);
  gluLookAt(viewer.Position().x, viewer.Position().y, viewer.Position().z,
            viewer.LookAt().x,   viewer.LookAt().y,   viewer.LookAt().z,
            viewer.Up().x,       viewer.Up().y,       viewer.Up().z);

  skyMap.Render(viewer.Position(), timeOfDay);
  Lighting::Do(timeOfDay);
  terrain.Render();
  glTranslatef(posHenge.x, posHenge.y, posHenge.z);
  glRotatef(-90, 0.0, 1.0, 0.0);
  mesh.Render();

  banner.Render() ;

  glFlush();
  glutSwapBuffers();
  glDisable(GL_TEXTURE_2D);
}

void init() {
  glClearColor (0.15, 0.15, 0.2, 0.0);
  Lighting::Setup();
  glEnable(GL_DEPTH_TEST);
}

void reshape(int w, int h) {
  glViewport(0, 0, (GLsizei) w, (GLsizei) h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(60.0, (GLfloat) w/(GLfloat) h, 1.0, 300.0);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glTranslatef(0.0, 0.0, -3.6);
}

// full screen toggle taken from
// http://www.york.ac.uk/services/cserv/sw/graphics/OPENGL/L24.c
void toggleFullscreen(const bool & full) {
  if (full) {
    Xpos = glutGet((GLenum)GLUT_WINDOW_X);
    Ypos = glutGet((GLenum)GLUT_WINDOW_Y);
    windowX = glutGet((GLenum)GLUT_WINDOW_WIDTH);
    windowY = glutGet((GLenum)GLUT_WINDOW_HEIGHT);
    glutFullScreen();
  } else {
    glutReshapeWindow(windowX, windowY);
    glutPositionWindow(Xpos, Ypos);
  }
}

void readKeyboard(unsigned char key, int x, int y) {startIdleFunc() ;};
void readKeyboard(int key, int x, int y) {startIdleFunc() ;};
void startIdleFunc() {
  if(viewer.IsNotActive()) {
    // stuffs going down start the clock!
    timLast = GetTickCount();
  }
  glutIdleFunc(idleFunc);
}
void readKeyboard() {
	// glut keyboard functions have numerous shortcomings
	// using Windows asynchronous key read gets round most of these

	keys[256] = 0;
	for (int x =0; x < 256; x++) {
		keys[x] = KEYDOWN(x);
		keys[256] |= keys[x];
	}
}

void idleFunc() {
	glutIdleFunc(NULL);
	// this function is toggled on when user action ends 
	// ie on key up or mouse up 
	// and continues while the scene is changing 
	// ie while there is some velocity
	// the function is toggled off when velocity drops to zero

	// determine the time since the last call
	DWORD now = GetTickCount();
	DWORD timeSlice = now - timLast;

/* control keys
                       |  shift       | ctrl |       
   ------+-------------+--------------+------+-------------
   up    | forward     | look up      | rise |   
   down  | reverse     | look down    | fall | 
   left  | turn left   | strafe left  |
   right | turn right  | strafe right |

   F1 - F4 toggle spotlights
   TAB - cycle through time of day
   1 - 5 select preset camera position
*/
	readKeyboard();
	viewer.Accelerate(0);
	viewer.Strafe(0);
	if(keys[256]) {
		//if(keys[VK_ESCAPE]) {
		//  banner.Enable(2);
		//  return;
		//}

		// debug tool, cycle through original normals and calculate flat and smooth
		if(keys['P']) {
			if(keys[VK_SHIFT])
				mesh.shadeStyle--;
			else
				mesh.shadeStyle++;
		}

		// cycle through the time of day
		if(keys[VK_TAB]) {
			if(timeOfDay == "day")
				timeOfDay = "twilight";
			else if(timeOfDay == "twilight")
				timeOfDay = "night";
			else if(timeOfDay == "night")
				timeOfDay = "day";
		}

		// handle motion
		if(keys[VK_SHIFT]) {
			if(keys[VK_CONTROL]) { // shift && control
			} else { // shift only
				if(keys[VK_LEFT]  && !keys[VK_RIGHT]) viewer.Strafe(-50);
				if(keys[VK_RIGHT] && !keys[VK_LEFT])  viewer.Strafe(50);
				if(keys[VK_UP]    && !keys[VK_DOWN])  viewer.Tilt(-5);
				if(keys[VK_DOWN]  && !keys[VK_UP])    viewer.Tilt(5);
			}
		} else {
			if(keys[VK_CONTROL]) { // control only
				if(keys[VK_UP]    && !keys[VK_DOWN])  viewer.ChangeHeightBy(0.1);
				if(keys[VK_DOWN]  && !keys[VK_UP])    viewer.ChangeHeightBy(-0.1);
			} else { // neither shift nor contorl
				if(keys[VK_LEFT]  && !keys[VK_RIGHT]) viewer.Turn(-5);
				if(keys[VK_RIGHT] && !keys[VK_LEFT])  viewer.Turn(5);
				if(keys[VK_UP]    && !keys[VK_DOWN])  viewer.Accelerate(50);
				if(keys[VK_DOWN]  && !keys[VK_UP])    viewer.Accelerate(-50);
			}
		}

		// toggle full screen
		if(keys[VK_F5]) {
			fullscreen = !fullscreen;
			toggleFullscreen(fullscreen);
		}

		// toggle spotlights
		if(keys[VK_F1]) {
			Lighting::lights[0] ^= true;
		}
		if(keys[VK_F2]) {
			Lighting::lights[1] ^= true;
		}
		if(keys[VK_F3]) {
			Lighting::lights[2] ^= true;
		}
		if(keys[VK_F4]) {
			Lighting::lights[3] ^= true;
		}

		// select camera position
		if(keys['1']) {
			viewer.TiltTo(cameraSetups[1].tilt);
			viewer.SetPosition(cameraSetups[1].position);
			viewer.RotateYto(cameraSetups[1].turn);
		} else if(keys['2']) {
			viewer.TiltTo(cameraSetups[2].tilt);
			viewer.SetPosition(cameraSetups[2].position);
			viewer.RotateYto(cameraSetups[2].turn);
		} else if(keys['3']) {
			viewer.TiltTo(cameraSetups[3].tilt);
			viewer.SetPosition(cameraSetups[3].position);
			viewer.RotateYto(cameraSetups[3].turn);
		} else if(keys['4']) {
			viewer.TiltTo(cameraSetups[4].tilt);
			viewer.SetPosition(cameraSetups[4].position);
			viewer.RotateYto(cameraSetups[4].turn);
		} else if(keys['5']) {
			viewer.TiltTo(cameraSetups[0].tilt);
			viewer.SetPosition(cameraSetups[0].position);
			viewer.RotateYto(cameraSetups[0].turn);
		} else if(keys['T']) {
			if(transition.target == -1) {
				// select a new camera position at random
				transition.target = (rand()%4);
				Vector3d t = cameraSetups[transition.target].position - viewer.Position();
				// if already at new camera try again, but only once, cos I fell mean
				if(Magnitude(t) < 0.1) {
					transition.target = (rand()%4);
					viewer.Position() - cameraSetups[transition.target].position;
				}
				// always travel at 0.1m/s during transition
				// therefore step size always = 0.1 
				// so get tenth of unit vector
				transition.stepSize = Normalize(t ) / 10;
			}
		}
	} 

	if(transition.target == -1) {
		viewer.Update(double(timeSlice)/1000);
		if(viewer.IsNotActive() && !banner.IsVisible()) {
			;//glutIdleFunc(NULL);
		} else {
			timLast = now;
			glutIdleFunc(idleFunc);
		}
	} else {
		viewer.SetPosition(viewer.Position() + transition.stepSize* double(timeSlice)/1000);
		
		Vector3d v = Vector3d( 50, 2, 50 ) - viewer.Position();
		viewer.LookAt(Normalize( v ));
		
		// has viewer reached position
		Vector3d t = viewer.Position() - cameraSetups[transition.target].position;
		double m = Magnitude(t);
		std::cout << m << std::endl;
		if(Magnitude(t) < 2.0) {
			viewer.TiltTo(cameraSetups[transition.target].tilt);
			viewer.SetPosition(cameraSetups[transition.target].position);
			viewer.RotateYto(cameraSetups[transition.target].turn);
			transition.target = -1;
		} else {
			glutIdleFunc(idleFunc);
		}
	}
	glutPostRedisplay();
}

int main(int argc, char** argv) {
#ifndef development
  if(MessageBox(NULL, "Would you like full screen?", "Window size", MB_YESNO | MB_ICONQUESTION) == IDYES) {
    fullscreen = true;
  }
#endif

  //////////////////////////////////////////////////////
  // set up the viewer's initial position and direction
  viewer.SetPosition(Vector3d(50,1.8,50));
  viewer.RotateYby(-70);
  Vector3d lowLim = Vector3d(2,0,2);
  Vector3d upLim = Vector3d(98,0,98);
  viewer.SetLimits(upLim, lowLim);

  MeshDataLoader::Load("sthenge3.obj", mesh);
  mesh.ScaleBy(0.115);
  mesh.FindBoundingBox();
  mesh.Centralise();
  mesh.FindBoundingBox();
  Vector3d ub = mesh.GetUpperBound();
  Vector3d lb = mesh.GetLowerBound();

  // set the transition target to -1 (none)
  transition.target = -1;

  // setup 5 different camera positions - found by trial and error
  cameraSetups[0].position = Vector3d(50,1.8,50);
  cameraSetups[0].tilt = 0;
  cameraSetups[0].turn = -70;

  cameraSetups[1].position = Vector3d(2,1.8,2);
  cameraSetups[1].tilt = 0;
  cameraSetups[1].turn = 45;

  cameraSetups[2].position = Vector3d(50,50,50);
  cameraSetups[2].tilt = 90;
  cameraSetups[2].turn = 0;

  cameraSetups[3].position = Vector3d(65,0,35);
  cameraSetups[3].tilt = -5;
  cameraSetups[3].turn = 130;

  cameraSetups[4].position = Vector3d(46,16,71);
  cameraSetups[4].tilt = 40;
  cameraSetups[4].turn = -75;

  // work out panoramic screen ratio 16:9
  unsigned int screenX = glutGet((GLenum)GLUT_SCREEN_WIDTH);
  unsigned int screenY = glutGet((GLenum)GLUT_SCREEN_HEIGHT);
  if(16 * screenY > 9 * screenX) {
    // X is small, so base ratio on that
    windowX = screenX;
    windowY = (9 * windowX) / 16;
  } else {
    // Y is small, so base ratio on that
    windowY = screenY;
    windowX = (16 * windowY) / 9;
  }
#ifdef development
    windowY = windowY * 0.75;
    windowX = windowX * 0.75;
#endif

  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGB | GLUT_DOUBLE | GLUT_DEPTH);
  glutInitWindowSize(windowX, windowY);
  glutInitWindowPosition(0, (screenY - windowY) /2);
  glutCreateWindow("Carl Bateman");

  if(fullscreen)
    toggleFullscreen(fullscreen);

#ifndef development
  timLast = GetTickCount();
  glutIdleFunc(idleFunc);
#endif

  init();
  glutDisplayFunc(display);
  glutReshapeFunc(reshape);

  glutKeyboardFunc(readKeyboard);
  glutKeyboardUpFunc(readKeyboard);
  glutSpecialFunc(readKeyboard);
  glutSpecialUpFunc(readKeyboard);

  skyMap.Init();
  terrain.InitTexture();

#ifndef development
  banner.Init();
  banner.Enable(0);
#endif

  glutMainLoop();
  return 0; 
}
